﻿using System;
using System.Collections.Generic;
using System.Data.Services.Providers;
using System.Reflection;
using System.Linq;

namespace CustomODataService.CustomDataServiceBase
{
	public class CustomDataServiceMetadataProvider : IDataServiceMetadataProvider
	{
		private Dictionary<string, ResourceType> resourceTypes = new Dictionary<string, ResourceType>();
		private Dictionary<string, ResourceSet> resourceSets = new Dictionary<string, ResourceSet>();

		/// <summary>
		/// Add a resource type
		/// </summary>
		/// <param name="type">Type to add</param>
		public void AddResourceType(ResourceType type)
		{
			type.SetReadOnly();
			resourceTypes.Add(type.FullName, type);
		}

		/// <summary>
		/// Adds the resource set.
		/// </summary>
		/// <param name="set">The set.</param>
		public void AddResourceSet(ResourceSet set)
		{
			set.SetReadOnly();
			resourceSets.Add(set.Name, set);
		}

		public static IDataServiceMetadataProvider BuildDefaultMetadataForClass<TEntity>(string namespaceName)
		{
			// Add resource type for class TEntity
			var productType = new ResourceType(
				typeof(TEntity),
				ResourceTypeKind.EntityType,
				null, // BaseType 
				namespaceName, // Namespace 
				typeof(TEntity).Name,
				false // Abstract? 
			);

			// use reflection to get all primitive properties
			foreach (var property in typeof(TEntity)
					.GetProperties(BindingFlags.Public | BindingFlags.Instance)
					.Where(pi => pi.DeclaringType == typeof(TEntity) && (pi.PropertyType.IsPrimitive || pi.PropertyType == typeof(string))))
			{
				var resourceProperty = new ResourceProperty(
					property.Name,
					// For simplicity let's assume that the property with postfix ID is the key property
					ResourcePropertyKind.Primitive | (property.Name.EndsWith("ID") ? ResourcePropertyKind.Key : 0),
					ResourceType.GetPrimitiveResourceType(property.PropertyType));
				productType.AddProperty(resourceProperty);
			}

			// Build metadata object
			var metadata = new CustomDataServiceMetadataProvider();
			metadata.AddResourceType(productType);
			metadata.AddResourceSet(new ResourceSet(typeof(TEntity).Name, productType));
			return metadata;
		}

		#region Implementation of IDataServiceMetadataProvider
		public string ContainerName
		{
			get { return "Container"; }
		}

		public string ContainerNamespace
		{
			get { return "Namespace"; }
		}

		public IEnumerable<ResourceType> GetDerivedTypes(ResourceType resourceType)
		{
			// We don't support type inheritance yet 
			yield break;
		}

		public ResourceAssociationSet GetResourceAssociationSet(ResourceSet resourceSet, ResourceType resourceType, ResourceProperty resourceProperty)
		{
			throw new NotImplementedException("No relationships.");
		}

		public bool HasDerivedTypes(ResourceType resourceType)
		{
			// We don’t support inheritance yet 
			return false;
		}

		public IEnumerable<ResourceSet> ResourceSets
		{
			get { return this.resourceSets.Values; }
		}

		public IEnumerable<ServiceOperation> ServiceOperations
		{
			// No service operations yet 
			get { yield break; }
		}

		public bool TryResolveResourceSet(string name, out ResourceSet resourceSet)
		{
			return resourceSets.TryGetValue(name, out resourceSet);
		}

		public bool TryResolveResourceType(string name, out ResourceType resourceType)
		{
			return resourceTypes.TryGetValue(name, out resourceType);
		}

		public bool TryResolveServiceOperation(string name, out ServiceOperation serviceOperation)
		{
			// No service operations are supported yet 
			serviceOperation = null;
			return false;
		}

		public IEnumerable<ResourceType> Types
		{
			get { return this.resourceTypes.Values; }
		}
		#endregion
	}
}